/*
GLUT program using GLSL to show how to grab rendered data into a texture.

Orion Sky Lawlor, olawlor@acm.org, 2006/10/05 (Public Domain)
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glew.h> /* GL Extensions Wrangler, http://glew.sourceforge.net/ */
#include <GL/glut.h> /* GL Utilities Toolkit, http://freeglut.sourceforge.net/ */
#include <GL/glui.h> /* GL User Interface, http://glui.sourceforge.net/ */
#include "osl/vec4.h" /* GLSL-style C++ types, by Orion Lawlor */
#include "osl/mat4.h"
using osl::mat4;

#include "ogl/program.h"
#include "ogl/program.cpp"
#include "ogl/util.cpp"
#include "ogl/fast_mipmaps.c"

#include "ogl/readback.h"
#include "ogl/readback.cpp"

#include "osl/random.h"
#include "osl/random_vector.h"
#include "osl/random.cpp" /* For easy linking */

#ifndef M_PI
# define M_PI           3.14159265358979323846  /* pi */
#endif


/* Variables updated by the GUI */
float cameraZoom=0.9, cameraDist=3.0;
int win_w, win_h; /**< Width and height (in pixels) of our window at last resize */
int view_w, view_h; /**< Width and height (in pixels) of our viewport at last resize */

/* These are rotation matrices, updated by the GLUI_Rotation spheres */
mat4 modelM(1.0), texM(1.0), viewM(1.0);
float modelMscale=1.0, texMscale=1.0, viewMscale=1.0;

float decayRate=1.0/255;
float growthRate=0.0;

/* Debugging details */
int detail1,detail2,detail3;

int modelElines=1, modelEbox=0;
int texE=1, texClamp=0;


void Exit(const char *where,const char *why) {
	fprintf (stderr, "FATALITY: OpenGL Error in %s: %s\n", where, why);
	abort();	
}

// Check GL for errors after the last command
void Check(const char *where) {
	GLenum e=glGetError();
	if (e==GL_NO_ERROR) return;
	const GLubyte *errString = gluErrorString(e);
	Exit(where, (const char *)errString);
}

/************* Objects we'll be drawing ***********/
class drawObject; /* forward declaration */

// Information about everything that can be drawn
class drawWorld {
public:
	int w,h; /* Size of texture */
	oglTexture *tex;
	
	
	
	drawWorld();
	void draw(void);
};

/*********************** DrawWorld *****************/
drawWorld::drawWorld() 
{
	w=h=512;
	tex=new oglTexture(NULL,w,h,oglTexture_linear);
}
/* Draw the vertices of a quad from -1 to +1 on all axes */
void drawQuadPoints(void) {
	oglVector3d pts[4]={ /* Points outlining a little quad */
		oglVector3d(-1,-1,0),oglVector3d(+1,-1,0),
		oglVector3d(+1,+1,0),oglVector3d(-1,+1,0)
	};
	int p,n=4;
	for (p=0;p<n;p++) glVertex3dv(pts[p]);
}

/* Draw a quad from -1 to +1 on all axes */
void drawQuad(void) {
	glBegin(GL_QUADS); /* Draw textured quad onscreen */
	drawQuadPoints();
	glEnd();
}

/* Make a 3D scaling matrix by this much */
mat4 makeScale(float by) {
	mat4 ret=mat4(by);
	ret[3].w=1.0; /* don't scale W coordinate, dangit! */
	return ret;
}

/** Main drawing routine, called by main_display below */
void drawWorld::draw(void)
{
	oglCheck("start draw");
	
	// Set up a little error-prone OpenGL state:
	glEnable(GL_BLEND); /* alpha blending */
	glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_ALPHA); /* premultiplied alpha */
	glEnable(GL_DEPTH_TEST);
	
/* First pass: draw something into our little texture */
	glViewport(0,0,w,h);
	tex->bind();
	GLenum wrapMode=GL_CLAMP;
	if (texClamp==1) wrapMode=GL_REPEAT;
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,wrapMode);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,wrapMode);
	
	glClearColor(0,0,0,0); /* set clear color--transparent */
	glClearDepth(1.0); /* set clear depth */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_ALPHA); /* GL_SRC_ALPHA */
	
	/* Blur the previous view of the world */
	if (texE) { 
		static oglProgramObject *blurP=oglProgramFromFiles("vertex.txt","fragment.txt");
		blurP->begin();
		blurP->set("modelM",modelM*makeScale(modelMscale));
		blurP->set("texM",texM*makeScale(texMscale));
		blurP->set("viewM",viewM*makeScale(viewMscale));
		blurP->set("pixel",vec4(1.0/w,1.0/h,0.0,0.0)); /* texture coordinate size of one pixel */
		blurP->set("decay",decayRate*vec4(1.0)); /* darkening rate */
		blurP->set("growth",growthRate*vec4(1.0)); /* growth rate */
		tex->bind();
		glDisable(GL_DEPTH_TEST);
		drawQuad();
		glEnable(GL_DEPTH_TEST);
		blurP->end();
	}
	
	/* Draw the model */
	if (modelElines || modelEbox) {
		static oglProgramObject *modelP=oglProgramFromStrings(
			"/* Vertex program */\n"
			"uniform mat4 modelM,texM,viewM;\n"
			"varying vec4 color;\n"
			"void main(void) { \n"
			"	gl_Position=modelM*gl_Vertex; \n"
			"	color=gl_Color;\n"
			"}\n"
		,
			"/* Fragment program-- just texture lookup */\n"
			"varying vec4 color;\n"
			"void main(void) { gl_FragColor=color; }\n"
		);
		modelP->begin();
		modelP->set("modelM",modelM*makeScale(modelMscale));
		modelP->set("texM",texM*makeScale(texMscale));
		modelP->set("viewM",viewM*makeScale(viewMscale));
		if (modelElines) {
			glBegin(GL_LINES);
			glColor4f(0,1,0,1); /* green horizontal slices */
			for (double dy=-0.9;dy<=+0.9;dy+=0.1) {
				glVertex2f(-0.5,dy);
				glVertex2f(+0.5,dy+0.03);
			}
			glColor4f(1,0,0,1); /* Red diagonal */
			glVertex2f(+0.7,-0.7);
			glVertex2f(-0.7,+0.7);
			glEnd();
		}
		if (modelEbox) {
			glBegin(GL_QUADS);
			glColor4f(0.5,0.5,0.6,1); /* light blue box */
			double s=0.1;
			double cx=0.6,cy=0.2;
			glVertex2f(cx-s,cy-s);
			glVertex2f(cx+s,cy-s);
			glVertex2f(cx+s,cy+s);
			glVertex2f(cx-s,cy+s);
			glEnd();
		}
		modelP->end();
	}
	
/* Read the stuff we just rendered back into our texture */
	tex->bind();
	glCopyTexSubImage2D(GL_TEXTURE_2D, /* target*/
		0, /* mipmap level */
		0,0, /* texture X and Y offset */
		0,0, /* onscreen X and Y offset */
		w,h  /* texture width and height */ 
	);
	
	if (detail1) return; /* just show first pass */
	
/* Set up screen for normal rendering */
	glViewport(0,0,view_w,view_h);
	// Clear the color and depth buffers
	glClearColor(0.2,0.3,0.4,1.0); /* set clear color */
	glClearDepth(1.0); /* set clear depth */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT); /* actually clear color and depth */
	
	// Perspective matrix
	double z=cameraZoom*cameraDist;
	mat4 persMat=mat4(1.0);
	persMat[0].x=z*view_h/(float)view_w; /* aspect ratio correction */
	persMat[1].y=z;
	persMat[2].z=-0.1; /* FIXME: set depth buffer range in some smarter way... */
	persMat[2].w=-1.0; /* Flip Z axis, so -Z is deeper into screen (bigger W) */
	persMat[3].w=cameraDist;
	
	// Overall coordinate transform: perspective  view   model
	mat4 clipMat=persMat*viewM*makeScale(viewMscale);
	
/* Set up pixel shader, and draw texture on a little plane */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	static oglProgramObject *simpleP=oglProgramFromStrings(
		"/* Vertex program */\n"
		"uniform mat4 clipFmWorld, worldFmModel;\n"
		"varying vec4 texCoords, color;\n"
		"void main(void) { \n"
		"	vec4 world=worldFmModel*gl_Vertex; /* world coords */ \n"
		"	gl_Position=clipFmWorld*world; \n"
		"	texCoords=0.5+0.5*gl_Vertex;\n"
		"	color=gl_Color;\n"
		"}\n"
	,
		"/* Fragment program-- just texture lookup */\n"
		"varying vec4 texCoords,color;\n"
		"uniform sampler2D tex;\n"
		"void main(void) { \n"
		"	vec4 T=texture2D(tex,vec2(texCoords));\n"
		"	gl_FragColor=T+color;\n"
		"}\n"
	);
	simpleP->begin();
	simpleP->set("clipFmWorld",clipMat); 
	simpleP->set("worldFmModel",mat4(1.0)); 
	simpleP->set("tex",0);
	oglCheck("after uploading simple program");
	
	tex->bind();
	glColor4f(0,0,0,0); /* <- let texture's colors shine through */
	/* Draw textured quad onscreen */
	drawQuad();

	glDisable(GL_DEPTH_TEST); /* otherwise quad and outline Z-fight */
	glColor4f(1,0,0,1); /* Draw a bright red outline */
	glBegin(GL_LINE_LOOP);
	drawQuadPoints();
	glEnd();
	glEnable(GL_DEPTH_TEST);
	
	simpleP->end();
}

/******** Program Initialization *********/
drawWorld *mainObject=0; /**< Big object we're trying to draw */

/* GLUT calls this routine every time our window needs to be redrawn */
void main_display(void) {
	Check("Entering display routine");
	
	// Draw stuff
	Check("About to draw");
	mainObject->draw();
	Check("Done drawing");
	
	// Display the stuff drawn so far to the screen:
	glutSwapBuffers();
	
	// This call makes us continually ask for another redraw
	glutPostRedisplay();
}

/* GLUT calls this routine every time our window changes size */
void main_reshape(int w, int h) {
	win_w=w, win_h=h;
	Check("Before glViewport");  
	int tx, ty, tw, th;
	GLUI_Master.get_viewport_area( &tx, &ty, &tw, &th );
	glViewport( tx, ty, view_w=tw, view_h=th );
	Check("After glViewport");
}

void make_list(GLUI *glui,const char *title,int *mode_ptr,const char **names)
{
	GLUI_Listbox *list = new GLUI_Listbox( glui, title, mode_ptr);
	for (int i=0;names[i];i++) 
		list->add_item(i,names[i]);
	list->do_selection(*mode_ptr);
}

int main(int argc, char** argv)
{
	glutInitWindowSize(900,750);
	glutInit(&argc,argv);

	// Ask for a back buffer, color buffer, and Z buffer in a window.
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH | GLUT_STENCIL;
	glutInitDisplayMode (mode);
	int main_window = glutCreateWindow("GLSL Lighting Demo");
	
	// Turn on OpenGL extensions, like programmable shaders
	//   WARNING: you MUST call glutCreateWindow before doing this!
	//   If you don't call glewInit, or call it wrong, you'll segfault!
	glewInit();
	
	// Make an object to draw
	mainObject=new drawWorld;
	
	// Make some sliders (this is the only GLUI code you'll need!)
	GLUI *glui = GLUI_Master.create_glui_subwindow(main_window, 
		GLUI_SUBWINDOW_TOP);
	glui->set_main_gfx_window( main_window );
	

	// Ask GLUT to call us when we need to redraw
	glutDisplayFunc (main_display);
	//glutReshapeFunc (main_reshape);
	GLUI_Master.set_glutReshapeFunc( main_reshape );  
	
	new GLUI_Checkbox(glui,"Show texture",&detail1);
	new GLUI_Checkbox(glui,"Detail 2:",&detail2);
	new GLUI_Checkbox(glui,"Detail 3:",&detail3);
	GLUI_Spinner *spin_d = new GLUI_Spinner( glui, "Decay Rate:", &decayRate);
	spin_d->set_float_limits(-0.1, 0.1);
	GLUI_Spinner *spin_g = new GLUI_Spinner( glui, "Growth Rate:", &growthRate);
	spin_g->set_float_limits(-0.1, 0.1);
	
	//new GLUI_Column(glui);
	//make_list(glui,"Model:",&model_mode,model_mode_names);
	
	
	new GLUI_Column(glui);
	new GLUI_Rotation(glui,"Model:",&modelM[0][0]);
	GLUI_Spinner *spin_a = new GLUI_Spinner( glui, "ModelScale:", &modelMscale);
	spin_a->set_float_limits(0.0, 10.0);
	new GLUI_Checkbox(glui,"Enable Lines",&modelElines);
	new GLUI_Checkbox(glui,"Enable Box",&modelEbox);
	//make_list(glui,"Type:",&model_mode,model_mode_names);
	//make_list(glui,"Animate:",&model_anim,model_anim_names);

	new GLUI_Column(glui,false);
	new GLUI_Rotation(glui,"Texture:",&texM[0][0]);
	GLUI_Spinner *spin_b = new GLUI_Spinner( glui, "TexScale:", &texMscale);
	spin_b->set_float_limits(0.9, 1.1);
	new GLUI_Checkbox(glui,"Enable Texture",&texE);
	new GLUI_Checkbox(glui,"Clamp Texcoords",&texClamp);
	
	
	new GLUI_Column(glui,false);
	new GLUI_Rotation(glui,"View:",&viewM[0][0]);
	GLUI_Spinner *spin_zoom = new GLUI_Spinner( glui, "Zoom:", &viewMscale);
	spin_zoom->set_float_limits(0.1, 10.0);
	GLUI_Spinner *spin_cd = new GLUI_Spinner( glui, "Distance:", &cameraDist);
	spin_cd->set_float_limits(0.1, 10.0);
	
	// Process events until user closes our window...
	glutMainLoop();
	
	return 0;
}
